/**
 * @file vga_text.c
 * VGA_Text native device driver interface.
 */
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <propeller.h>

#include "vga_asm.h"
#include "vga_text.h"
#include "VGA_Array.h"

/**
 * This is the main global vga text control/status structure.
 */
vgaText_t gVgaText;

/**
 * This is the VGA text screen area.
 */
static short gscreen[VGA_TEXT_SCREENSIZE];

/**
 * This is the VGA color palette area.
 */
static long gcolors[VGA_TEXT_COLORTABLE_SIZE+2];

/**
 * These are variables to keep up with display;
 */
static int col, row, flag;

static int blank = 0x220;

/**
 * This is the VGA palette.
 */
static char gpalette[VGA_TEXT_COLORTABLE_SIZE] =     
{                           // fgRGB  bgRGB    '
    0b111111, 0b000001,     // %%333, %%001    '0    white / dark blue
    0b111100, 0b010100,     // %%330, %%110    '1   yellow / brown
    0b100010, 0b000000,     // %%202, %%000    '2  magenta / black
    0b010101, 0b111111,     // %%111, %%333    '3     grey / white
    0b001111, 0b000101,     // %%033, %%011    '4     cyan / dark cyan
    0b001000, 0b101110,     // %%020, %%232    '5    green / gray-green
    0b010000, 0b110101,     // %%100, %%311    '6      red / pink
    0b001111, 0b000011      // %%033, %%003    '7     cyan / blue
};

/*
 * This should set the character foreground and screen background.
 * API are available for this.
 */
static int color = 0;

/*
 * delay function used after start cog and during screen updates
 * @param per is number of ticks to wait ....
 * Careful, if per is too small, the processor will wait until the next count
 * which can be on the order of minutes. 
 */
static void wait(int per)
{
    int time = 0;
    asm("mov        %time, cnt");
    asm("add        %time, %per");
    asm("waitcnt    %time, %per");
}    

/*
 * VGA_Text start function starts VGA on a cog
 * See header file for more details.
 */
int     vgaText_start(int basepin)
{
    int id = 0;

    col   = 0; // init vars
    row   = 0;
    flag  = 0;

    gVgaText.status = 0;
    gVgaText.enable = 1;
    gVgaText.pins   = basepin | 0x7;
    gVgaText.mode   = 0b1000;
    gVgaText.screen = (long) gscreen;
    gVgaText.colors = (long) gcolors;
    gVgaText.ht = VGA_TEXT_COLS;
    gVgaText.vt = VGA_TEXT_ROWS;
    gVgaText.hx = 1;
    gVgaText.vx = 1;
    gVgaText.ho = 0;
    gVgaText.vo = 0;
    gVgaText.hd = 512;
    gVgaText.hf = 10;
    gVgaText.hs = 75;
    gVgaText.hb = 43;
    gVgaText.vd = 480;
    gVgaText.vf = 11;
    gVgaText.vs = 2;
    gVgaText.vb = 31;
    gVgaText.rate = 80000000 >> 2;
      
    id = cognewNative(VGA_Text_Array, &gVgaText);

    // set main fg/bg color here
    vgaText_setColorPalette(&gpalette[VGA_TEXT_PAL_WHITE_BLUE]);
    vgaText_out(0);
    
    return id;
}

/*
 * VGA_Text stop function stops VGA cog
 * See header file for more details.
 */
void    vgaText_stop(int id)
{
    if(id) {
        asm("cogstop %id");
    }
}

/*
 * VGA_Text setcolors function sets the palette to that defined by pointer.
 * See header file for more details.
 */
void    vgaText_setColorPalette(char* ptr)
{
    int  ii = 0;
    int  mm = 0;
    char fg = 0;
    char bg = 0;
    for(ii = 0; ii < VGA_TEXT_COLORTABLE_SIZE; ii += 2)
    {
        mm = mm + 1;    // beta version has trouble with math in braces. use mm
        fg = ptr[ii] << 2;
        bg = ptr[mm] << 2;
        gcolors[ii]  = fg << 24 | bg << 16 | fg << 8 | bg;
        gcolors[mm]  = fg << 24 | fg << 16 | bg << 8 | bg;
   }        
}

/*
 * print a new line
 */
static void newline(void)
{
    col = 0;
    if (++row == VGA_TEXT_ROWS) {
        row--;
        wordmove(&gscreen[0], &gscreen[VGA_TEXT_COLS], VGA_TEXT_LASTROW); // scroll
        wordfill(&gscreen[VGA_TEXT_LASTROW], blank, VGA_TEXT_COLS); // clear new line
    }
}

/*
 * print a character
 */
static void printc(int c)
{
    int   ndx = row * VGA_TEXT_COLS + col;
    short val = 0;
    
    // I can't seem to get the palette colors to work correctly.
    // If you want just one screen fg/bg color, use setColorPalette(&gcolors[N]);
    // It seems the color would be wrong for every "even" char.
    // Fine but take out the (c & 1) and the character is wrong.
    
    val  = (color << 1 | c & 1) << 10;
    val += 0x200 + (c & 0xFE);
    
    // Driver updates during invisible. Need some delay so screen updates right.
    // For flicker-free once per scan update, you can wait for status != invisible.
    // while(gVgaText.status != VGA_TEXT_STAT_INVISIBLE)    ;
    
    // Use some delay so printing works correctly.
    wait(50);
    
    gscreen[ndx] = val; // works
    //gscreen[row * VGA_TEXT_COLS + col] = val; // fails ... don't know why

    if (++col == VGA_TEXT_COLS) {
        newline();
    }
}


/*
 * VGA_Text str function prints a string at current position
 * See header file for more details.
 */
void    vgaText_str(char* sptr)
{
    while(*sptr) {
        vgaText_out(*(sptr++));
    }
}

/**
 * Reverse contents of a string.
 * @param s is string to reverse.
 * @returns reverse of s.
 */
static char* reverse(char* s)
{
    char* r = s+strlen(s)-1;
    char  ch;
    while(s < r) {
        ch = *s;
        *(s++) = *r;
        *(r--) = ch;
    }
    return r;
}

/**
 * Convert ascii string to signed integer.
 * @param s ascii string
 * @returns integer representation of string s.
 */
static int aton(char* s)
{
    int num  = 0;
    int sign = 1;
    while(isspace(*s))
        s++;
    if(*s == '-') {
        sign = -1;
        s++;
    }
    else if(*s == '+') {
        s++;
    }
    while(isdigit(*s)) {
        num = 10*num + *(s++) - '0';
    }
    return sign * num;
}

/**
 * Convert number to an ascii string. BETA library itoa is broken.
 * @param num is number.
 * @param s is string space for ascii string representation of integer num.
 * @param radix is radix of the number.
 */
static void ntoa(char* s, int num, int radix)
{
    char* start = s;
    int lowbit;
    radix = (radix >> 1) & 0xf;
    do {
        lowbit = num & 1;
        num = (num >> 1); // don't need mask
        *s = (mod(num, radix) << 1) + lowbit; // % operator appears to be broken in BETA!
        if(*s < 10)
            *s += '0';
        else
            *s += ('A'-10);
        s++;
        num = num / radix;
    } while (num != 0);
    *s = '\0';
    s = reverse(start);
}

/*
 * VGA_Text dec function prints a decimal number at current position
 * See header file for more details.
 */
void    vgaText_dec(int value)
{
    char b[128];
    ntoa(b, value, 10);
    vgaText_str(b);
}

/*
 * VGA_Text hex function prints a hexadecimal number at current position
 * See header file for more details.
 */
void    vgaText_hex(int value, int digits)
{
    int ndx;
    char hexlookup[] =
    {
        '0', '1', '2', '3', '4', '5', '6', '7',
        '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
    };
    while(digits-- > 0) {
        ndx = (value >> (digits<<2)) & 0xf;
        printc(hexlookup[ndx]);
    }   
}


/*
 * VGA_Text bin function prints a binary number at current position
 * See header file for more details.
 */
void    vgaText_bin(int value, int digits)
{
    int bit = 0;
    while(digits-- > 0) {
        bit = (value >> digits) & 1;
        printc(bit + '0');
    }   
}


/*
 * VGA_Text out function prints a character at current position or performs
 * a screen function.
 * See header file for more details.
 */
void    vgaText_out(int c)
{
    if(flag == 0)
    {
        switch(c)
        {
            case 0:
                wordfill(&gscreen[0], color << 11 | blank, VGA_TEXT_SCREENSIZE);
                col = 0;
                row = 0;
                break;
            case 1:
                col = 0;
                row = 0;
                break;
            case 8:
                if (col)
                    col--;
                break;
            case 9:
                do {
                    printc(' ');
                } while(col & 7);
                break;
            case 0xA:   // fall though
            case 0xB:   // fall though
            case 0xC:   // fall though
                flag = c;
                return;
            case 0xD:
                newline();
                break;
            default:
                printc(c);
                break;
        }
    }
    else
    if (flag == 0xA) {
        col = c % VGA_TEXT_COLS;
    }
    else
    if (flag == 0xB) {
        row = c % VGA_TEXT_ROWS;
    }
    else
    if (flag == 0xC) {
        color = c & 0xf;
    }
    flag = 0;
}

/*
 * VGA_Text print null terminated char* to screen with normal stdio definitions
 * See header file for more details.
 */
void    vgaText_print(char* s)
{
    while(*s) {
        vgaText_putchar(*(s++));
    }
}

/*
 * VGA_Text putchar print char to screen with normal stdio definitions
 * See header file for more details.
 */
void    vgaText_putchar(char c)
{
    switch(c)
    {
        case '\b':
            if (col)
                col--;
            break;
        case '\t':
            do {
                printc(' ');
            } while(col & 7);
            break;
        case '\n':
            newline();
            break;
        case '\r':
            col = 0;
            break;
        default:
            printc(c);
            break;
    }
}

/*
 * VGA_Text setCurPosition function sets position to x,y.
 * See header file for more details.
 */
void    vgaText_setCurPosition(int x, int y)
{
    col = x;
    row = y;
}

/*
 * VGA_Text setCoordPosition function sets position to Cartesian x,y.
 * See header file for more details.
 */
void    vgaText_setCoordPosition(int x, int y)
{
    col = x;
    row = VGA_TEXT_ROWS-y-1;
}

/*
 * VGA_Text setXY function sets position to x,y.
 * See header file for more details.
 */
void    vgaText_setXY(int x, int y)
{
    col = x;
    row = y;
}

/*
 * VGA_Text setX function sets column position value
 * See header file for more details.
 */
void    vgaText_setX(int value)
{
    col = value;
}

/*
 * VGA_Text setY function sets row position value
 * See header file for more details.
 */
void    vgaText_setY(int value)
{
    row = value;
}

/*
 * VGA_Text getX function gets column position
 * See header file for more details.
 */
int vgaText_getX(void)
{
    return col;
}

/*
 * VGA_Text getY function gets row position
 * See header file for more details.
 */
int vgaText_getY(void)
{
    return row;
}

/*
 * VGA_Text setColors function sets palette color set index
 * See header file for more details.
 */
void vgaText_setColors(int value)
{
    color = value;
}

/*
 * VGA_Text getColors function gets palette color set index
 * See header file for more details.
 */
int vgaText_getColors(void)
{
    return color;
}


#ifdef  VGA_TEXT_TEST
void main(void)
{
    int ii = 85;
    int jj = 0;

    vgaText_start(16);  // start VGA
    
    vgaText_out(0);
    vgaText_out(0xd);
    vgaText_str("zoot ");

    vgaText_str("color");
    vgaText_out(1);
    vgaText_str("Hi");
    vgaText_out(0xd);
    vgaText_str("zoot");
    vgaText_setCoordPosition(1,3);
    vgaText_str("abcd");
    vgaText_setCurPosition(5,5);
    
    vgaText_str("Hello World!\r");
    vgaText_setY(vgaText_getY()+1);
    vgaText_str("Decimal ");
    vgaText_dec(ii);
    vgaText_setY(vgaText_getY()+1);

    for(jj = 10; jj < 20; jj++) {
        vgaText_setXY(jj,mod(jj,VGA_TEXT_ROWS));
        vgaText_str("0x");
        vgaText_hex(ii++, 4);
    }
    
    while(1);
}
#endif
// VGA_TEXT_TEST
